/* $Id: catedit.c,v 1.5 2004/11/16 21:51:36 mjhammel Exp $ */

/*========================================================================
 *  Copyright (c) Michael J. Hammel 1995-2004
 *========================================================================
 *              FILE NAME: catedit.c 
 *            DESCRIPTION: category editor for all Pilot managers.
 *      DEFINED CONSTANTS: 
 *       TYPE DEFINITIONS: 
 *      MACRO DEFINITIONS: 
 *       GLOBAL VARIABLES: 
 *       PUBLIC FUNCTIONS: XPCatEdit
 *      PRIVATE FUNCTIONS: 
 *  SOFTWARE DEPENDENCIES: 
 *  HARDWARE DEPENDENCIES: 
 *                  NOTES: 
 *
 * SPECIAL CONSIDERATIONS:
 *========================================================================
 *
 * MODIFICATION HISTORY:
 * $Log: catedit.c,v $
 * Revision 1.5  2004/11/16 21:51:36  mjhammel
 * Category editing is now complete for Todos.
 * catedit.c complete for generic use by other Pilot managers.
 *
 * Revision 1.4  2004/11/16 19:06:38  mjhammel
 * Removed test for selected row in CatEditAdd() when adding a new category.
 *
 * Revision 1.3  2004/11/16 19:04:01  mjhammel
 * Added callbacks to postprocess category display options after category
 * edits have been made.
 *
 * Revision 1.2  2004/11/15 23:53:13  mjhammel
 * First working version of category editor.
 *
 * Revision 1.1  2004/11/15 20:29:27  mjhammel
 * Added first pass at Category Edit dialog.
 *
 *
 *========================================================================*/
#define XP_CATEDIT_C

/* === System Headers === */
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <gtk/gtk.h>
#include <gdk/gdkx.h>
#include <gdk/gdkkeysyms.h>


/* === Project Headers === */
#include "XNotesPlus.h"

/* Used to pass user data to the Modify callback. */
typedef struct _moddata_t {
	PilotCategory_T	**hdr;
	GtkWidget			*parent;
	void					(*updatefunc)();
} ModData_T;


/* === external routines === */


/* === Public routine prototypes === */


/* === Private routine prototypes === */
static void CatEditClose();
static void CatSelected();
static void CatEditModify();
static void CatEditAdd();
static void CatEditDelete();
static void UpdateDisplay();


/* === Global Variables === */
extern GtkWidget			*toplevel;
extern PilotCategory_T	*Pilot_Todo_Categories;

static GtkWidget	*cateditwidget = NULL;
static GtkWidget	*cat_text = NULL;
static GtkWidget	*cat_list = NULL;
static int			active_row = -1;
static int			destroy_id;



/*========================================================================
 *	Name:			XPCatEdit
 *	Prototype:	XPCatEdit(PilotCategory_T *)
 *					
 *	Description:
 * Edit categories (add/delete/update) for all Pilot managers
 *
 *	Input Arguments:
 *	PilotCategory_T **hdr	head of link list to update.
 *	GtkWidget	*parent		calling dialog, to make insensitive temporarily.
 *	gpointer		updatefunc	Callback when mods are made to existing categories.
 *	gpointer		updatefunc	Callback right before this dialog closes.
 *
 *	Output Arguments:
 *	Return Values:
 *	Global Variables:
 *	External Routines:
 *	Method:
 *	Restrictions:
 *	Notes:
 *========================================================================*/
void
XPCatEdit(
	PilotCategory_T	**hdr,
	GtkWidget			*parent,
	gpointer				updatefunc,
	gpointer				finishfunc
)
{
	GtkWidget		*shell;
	GtkWidget		*cancel_button;
	GtkWidget		*vbox, *hbox, *label;
	GtkWidget		*cat_viewport;
	GtkWidget		*delete_cat;
	GtkWidget		*modify_cat;
	GtkWidget		*add_cat;
	GShell_T			gshell;

	PilotCategory_T		*type = *hdr;
	static ModData_T		moddata;
	static ModData_T		findata;

	/* Don't allow multiple versions of this dialog at one time. */
	if ( cateditwidget != NULL )
		return;


	/* Make the parent insensitive while we do our stuff. */
	gtk_widget_set_sensitive ( GTK_WIDGET(parent), FALSE );

	/*
	 * Save the cb func and list ptr for use in our callbacks.
	 */
	moddata.hdr = hdr;
	moddata.updatefunc = updatefunc;
	findata.parent = NULL;
	findata.hdr = hdr;
	findata.updatefunc = finishfunc;
	findata.parent = parent;

	/*
	 * Create a generic dialog.
	 */
	cateditwidget = XPCreateGenericShell(CategoryEditDialog,
			CANCEL_BUTTON, &gshell, FALSE );

	destroy_id = gtk_signal_connect (GTK_OBJECT (cateditwidget), "destroy",
			GTK_SIGNAL_FUNC (CatEditClose), (gpointer)&findata);

	shell = gshell.vbox;
	cancel_button = gshell.cancel;
	gtk_label_set_text(GTK_LABEL(GTK_BIN(cancel_button)->child),XP_S_DONE);

	/*
	 * The dialog has two panes:
	 * 1. Left side:  the list of categories.
	 * 2. Right side: editing options.
	 */
	hbox = gtk_hbox_new(FALSE, 0);
	gtk_container_set_border_width (GTK_CONTAINER (hbox), 0);
	gtk_box_pack_start(GTK_BOX(shell), hbox, FALSE, TRUE, 0);
	gtk_widget_show(hbox);

	/*
	 * Left pane:  scrolled list of categories.
	 * Note:  There will never be 1000's or even 100's of
	 * categories, so forcing the user to delete one
	 * category at a time is not unreasonable.  Adding
	 * and modifying are, by their nature, one-at-a-time
	 * operations anyway.
	 */
	cat_viewport = gtk_scrolled_window_new(NULL, NULL);
	gtk_widget_set_name (cat_viewport, "cat_list");
	gtk_scrolled_window_set_policy (
		GTK_SCROLLED_WINDOW (cat_viewport),
		GTK_POLICY_NEVER,
		GTK_POLICY_AUTOMATIC);
	gtk_box_pack_start (GTK_BOX (hbox), cat_viewport, TRUE, TRUE, 0);
	gtk_widget_show (cat_viewport);

	cat_list = gtk_clist_new(1);
	gtk_widget_set_name (cat_list, "cat_list");
	gtk_clist_set_selection_mode (GTK_CLIST (cat_list), GTK_SELECTION_SINGLE);
	gtk_container_add(GTK_CONTAINER(cat_viewport), cat_list);
	gtk_widget_show (cat_list);
	gtk_clist_set_column_auto_resize (GTK_CLIST(cat_list), 0, TRUE);

	/*
	 * Right pane:  vertical box of options.
	 * 1. Delete selected category button.
	 * 2. Text entry box.
	 * 3. Modify Entry button.
	 * 4. Add entry button.
	 */
	vbox = gtk_vbox_new(FALSE, 1);
	gtk_container_set_border_width (GTK_CONTAINER (vbox), 0);
	gtk_container_add(GTK_CONTAINER(hbox), vbox);
	gtk_widget_show(vbox);

	/* The delete button and tooltip */
	delete_cat = gtk_button_new();
	gtk_widget_set_name(delete_cat, "window_button");
	label = gtk_label_new(XP_S_CAT_DELETE);
	gtk_widget_set_name(label, "window_button");
	gtk_widget_show(label);
	gtk_container_add(GTK_CONTAINER(delete_cat), label);
	gtk_widget_show(delete_cat);
	gtk_box_pack_start (GTK_BOX (vbox), delete_cat, FALSE, FALSE, 5);

	gtk_signal_connect(GTK_OBJECT(delete_cat), "event",
		GTK_SIGNAL_FUNC (XPToolTips), (gpointer)XP_TTID_CAT_DELETE);

	/* The text input/edit field. */
	cat_text = gtk_entry_new();
	gtk_widget_show(cat_text);
	gtk_container_add(GTK_CONTAINER(vbox), cat_text);

	/* The modify button and tooltip */
	modify_cat = gtk_button_new();
	gtk_widget_set_name(modify_cat, "window_button");
	label = gtk_label_new(XP_S_CAT_MODIFY);
	gtk_widget_set_name(label, "window_button");
	gtk_widget_show(label);
	gtk_container_add(GTK_CONTAINER(modify_cat), label);
	gtk_widget_show(modify_cat);
	gtk_box_pack_start (GTK_BOX (vbox), modify_cat, FALSE, FALSE, 5);

	gtk_signal_connect(GTK_OBJECT(modify_cat), "event",
		GTK_SIGNAL_FUNC (XPToolTips), (gpointer)XP_TTID_CAT_MODIFY);

	/* The add button and tooltip */
	add_cat = gtk_button_new();
	gtk_widget_set_name(add_cat, "window_button");
	label = gtk_label_new(XP_S_CAT_ADD);
	gtk_widget_set_name(label, "window_button");
	gtk_widget_show(label);
	gtk_container_add(GTK_CONTAINER(add_cat), label);
	gtk_widget_show(add_cat);
	gtk_box_pack_start (GTK_BOX (vbox), add_cat, FALSE, FALSE, 5);

	gtk_signal_connect(GTK_OBJECT(add_cat), "event",
		GTK_SIGNAL_FUNC (XPToolTips), (gpointer)XP_TTID_CAT_ADD);

	/*
	 * Cancellation button.
	 */
	gtk_signal_connect (GTK_OBJECT (cancel_button), "clicked",
			GTK_SIGNAL_FUNC (CatEditClose), (gpointer)&findata);

	/*
	 * Modify button.
	 */
	gtk_signal_connect (GTK_OBJECT (modify_cat), "clicked",
			GTK_SIGNAL_FUNC (CatEditModify), (gpointer)&moddata);

	/*
	 * Add button.
	 */
	gtk_signal_connect (GTK_OBJECT (add_cat), "clicked",
			GTK_SIGNAL_FUNC (CatEditAdd), (gpointer)type);

	/*
	 * Delete button.
	 */
	gtk_signal_connect (GTK_OBJECT (delete_cat), "clicked",
			GTK_SIGNAL_FUNC (CatEditDelete), (gpointer)&moddata);


	/*
	 * Populate the list of categories.
	 */
	UpdateDisplay(type);
	active_row = -1;

	/* When a category is selected, update the dialog. */
	gtk_signal_connect (GTK_OBJECT (cat_list), "select_row",
		GTK_SIGNAL_FUNC (CatSelected), (gpointer) type);

	/*
	 * Prevent resizing.
	 */
	gtk_window_set_policy(GTK_WINDOW(cateditwidget), FALSE, FALSE, FALSE);

	/*
	 * Pop up the confirmation box.
	 */
	gtk_widget_realize(cateditwidget);
	XPCenterWindow(cateditwidget, NULL, cateditwidget);
	gtk_widget_show( cateditwidget );

}

/*========================================================================
 *	Name:			CatEditClose
 *					
 * Input Arguments
 * ModData_T			*data		type: unused
 *										updatefunc: points to cb function from caller.
 *										parent: parent window to set sensitive.
 *
 *	Description:
 * Clean up and close the dialog.
 *========================================================================*/
void
CatEditClose(
	GtkWidget			*widget,
	ModData_T			*data
)
{
	if ( cateditwidget != NULL )
	{
		/* Block the destroy handler or we get called twice. */
		gtk_signal_handler_block (GTK_OBJECT (cateditwidget), destroy_id);
		gtk_widget_destroy( cateditwidget );
		cateditwidget = NULL;

		/* Make the parent allow input again. */
		gtk_widget_set_sensitive ( GTK_WIDGET(data->parent), TRUE );

		/* Let the calling manager update their display first. */
		(*data->updatefunc)();
	}
}

/*========================================================================
 *	Name:			CatSelected
 *					
 *	Description:
 * Update the text input field when a category is chosen from the list.
 *========================================================================*/
void
CatSelected(
	GtkWidget			*widget,
	gint					row,
	gint					column,
	GdkEventButton		*event,
	PilotCategory_T	*type
)
{
	PilotCategory_T	*cat_link;

	cat_link = gtk_clist_get_row_data(GTK_CLIST(widget), row);
	gtk_entry_set_text(GTK_ENTRY(cat_text), cat_link->name);
	active_row = row;
}

/*========================================================================
 *	Name:			CatEditModify
 *					
 * Input Arguments
 * ModData_T			*data		type: points to PilotCategory_T list
 *										updatefunc: points to cb function from caller.
 *										parent: unused.
 *
 *	Description:
 * Modify an existing entry in the list.
 *========================================================================*/
void
CatEditModify(
	GtkWidget			*widget,
	ModData_T			*data
)
{
	PilotCategory_T	*cat_link;
	char	*name;

	/* Don't do anything if the active row has not been set yet. */
	if ( active_row == -1 ) return;

	name = gtk_editable_get_chars(GTK_EDITABLE(cat_text), 0, -1);
	cat_link = gtk_clist_get_row_data(GTK_CLIST(cat_list), active_row);

	/* Let the calling manager update their lists first */
	(*data->updatefunc)(cat_link->name, name);

	/* Now update the category list */
	if ( cat_link->name ) free (cat_link->name);
	cat_link->name = XPSafeAlloc(strlen(name)+1);
	strcpy(cat_link->name, name);
	UpdateDisplay(*(data->hdr));
}

/*========================================================================
 *	Name:			CatEditAdd
 *					
 * Input Arguments
 *	PilotCategory_T	*type			Defines which link list we need to update.
 *
 *	Description:
 * Add an entry to the list.
 *========================================================================*/
void
CatEditAdd(
	GtkWidget			*widget,
	PilotCategory_T	*type
)
{
	char	*name;

	name = gtk_editable_get_chars(GTK_EDITABLE(cat_text), 0, -1);
	XPGenerateCategoryList(name, &Pilot_Todo_Categories);
	UpdateDisplay(type);
}

/*========================================================================
 *	Name:			CatEditDelete
 *					
 * Input Arguments
 * ModData_T			*data		type: points to PilotCategory_T list
 *										updatefunc: points to cb function from caller.
 *										parent: unused.
 *
 *	Description:
 * Delete an entry from the list.
 *========================================================================*/
void
CatEditDelete(
	GtkWidget			*widget,
	ModData_T			*data
)
{
	PilotCategory_T	*cat_link;
	PilotCategory_T	*searchlink;
	PilotCategory_T	*lastlink;

	/* Don't do anything if the active row has not been set yet. */
	if ( active_row == -1 ) return;

	cat_link = gtk_clist_get_row_data(GTK_CLIST(cat_list), active_row);
	for(searchlink=*(data->hdr); searchlink!=NULL; searchlink=searchlink->next)
	{
		if (searchlink->next == cat_link)
		{
			searchlink->next = cat_link->next;
			break;
		}
	}
	if ( searchlink == NULL ) return;

	/* Let the calling manager update their lists first */
	(*data->updatefunc)(cat_link->name, XP_S_CAT_UNFILED);

	/* Remove the link from the list. */
	lastlink = NULL;
	searchlink = *(data->hdr);
	while( searchlink != NULL )
	{
		if ( strcmp(searchlink->name, cat_link->name) == 0 )
		{
			free(searchlink->name);
			if ( lastlink == NULL )
				*(data->hdr) = searchlink->next;
			else
				lastlink->next = searchlink->next;
			free(searchlink);
			break;
		}

		lastlink = searchlink;
		searchlink = searchlink->next;
	}
	
	UpdateDisplay(*(data->hdr));
}

/*========================================================================
 *	Name:			UpdateDisplay
 *					
 * Input Arguments
 *	PilotCategory_T	*type			Defines which link list we are reading from.
 *
 *	Description:
 * Update the list of categories in the dialog.
 *========================================================================*/
void
UpdateDisplay(
	PilotCategory_T	*type
)
{
	char	row_data[1][512];
	char	*rows[1];
	int	row;
	int	index;
                                                                                
	PilotCategory_T	*cat_link;

	gtk_clist_freeze(GTK_CLIST(cat_list));
	gtk_clist_clear(GTK_CLIST(cat_list));
	cat_link = type;
	index = 0;
	while ( cat_link != NULL )
	{
		sprintf((char *)row_data[0], cat_link->name);
		rows[0] = row_data[0];
		row = gtk_clist_append(GTK_CLIST(cat_list), rows);
		gtk_clist_set_row_data(GTK_CLIST(cat_list), 
			index, (gpointer)cat_link);

		cat_link = cat_link->next;
		index++;
	}
	gtk_clist_thaw(GTK_CLIST(cat_list));
	gtk_entry_set_text(GTK_ENTRY(cat_text), "");

	/* Deselect everything */
	gtk_clist_unselect_all(GTK_CLIST(cat_list));
	active_row = -1;
}

