/* $Id: findtext.c,v 1.4 2005/07/05 21:37:34 mjhammel Exp $ */

/*========================================================================
 *  Copyright (c) Michael J. Hammel 1995-2001
 *========================================================================
 *              FILE NAME: findtext.c 
 *            DESCRIPTION: Find a text substring within a note
 *      DEFINED CONSTANTS: 
 *       TYPE DEFINITIONS: 
 *      MACRO DEFINITIONS: 
 *       GLOBAL VARIABLES: 
 *       PUBLIC FUNCTIONS: XPFindTextPrompt, XPFind
 *      PRIVATE FUNCTIONS: SearchIt, ClearFind, FindForward, FindBackward
 *  SOFTWARE DEPENDENCIES: 
 *  HARDWARE DEPENDENCIES: 
 *                  NOTES: 
 *
 * SPECIAL CONSIDERATIONS:
 * Set your tabstops to 3 to make the code more readable.
 *========================================================================*/
#define XP_FINDTEXT_C

/* === System Headers === */
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include <gtk/gtk.h>
#include <gdk/gdkkeysyms.h>


/* === Project Headers === */
#include "XNotesPlus.h"
#include "debug.h"


/* === external routines === */


/* === Public routine prototypes === */


/* === Private routine prototypes === */
static void		ClearFind();
static void		SearchIt();
static void		FindHome();
static gboolean	AKeyPress();


/* === Global Variables === */

/* === Static Variables === */
static GtkWidget		*findwidget;
static GtkWidget		*case_toggle, *direction_toggle;
static GtkWidget		*search_term;
static gboolean		firsttime;
static GtkTextMark	*home;

static int	start, end;
static char	*copy=NULL;



/*========================================================================
 *	Name:			XPFindTextPrompt
 *	Prototype:	XPFindTextPrompt()
 *					
 *	Description:
 * Put up a window that says who I am and what this thing is.
 *
 *	Input Arguments:
 *	Output Arguments:
 *	Return Values:
 *	Global Variables:
 *	External Routines:
 *	Method:
 *	Restrictions:
 *	Notes:
 *========================================================================*/
void
XPFindTextPrompt(
	int			index
)
{
#ifdef DEBUG
	char	fname[]="XPFindTextPrompt";
#endif

	GtkWidget		*vbox, *hbox;
	GtkWidget		*cancel_button, *accept_button, *home_button;
	GtkWidget		*label;
	GShell_T			gshell;
	StickyNote		*pn;

	firsttime = TRUE;

	/*
	 * Allow only one copy at a time.
	 */
	if ( findwidget != NULL )
		return;

	/*
	 * Grab the note structure.
	 */
	if ((pn = XPFindNote(index)) == NULL)
		return;

	/*
	 * Create a dialog, set its title, and position it at the cursor.
	 */
	DBGPrintf(DBG_INFO,("Creating find dialog.\n"));
	findwidget = XPCreateGenericShell(FindTextDialog, 
						SEARCH_BUTTON | CLOSE_BUTTON | OVERWRITE_BUTTON, 
						&gshell, FALSE );

	gtk_signal_connect (GTK_OBJECT (findwidget), "destroy",
			GTK_SIGNAL_FUNC (ClearFind), pn);

	/*
	 * Get the widget ids of the various parts of the window.
	 */
	vbox = gshell.vbox;
	cancel_button = gshell.cancel;
	accept_button = gshell.search;
	home_button   = gshell.overwrite;

	/*
	 * Change the label for the Home button.
	 */
	gtk_label_set_text(GTK_LABEL(GTK_BIN(home_button)->child),XP_S_HOME);

	/*
	 * A label describing what this dialog is for.
	 */
	label = gtk_label_new(XP_S_SEARCH_PROMPT);
	gtk_widget_show(label);
	gtk_container_add(GTK_CONTAINER(vbox), label);

	/*
	 * An text entry widget in which the search term will go.
	 */
	search_term = gtk_entry_new();
	gtk_widget_set_name (search_term, "find_list");
	gtk_widget_show(search_term);
	gtk_container_add(GTK_CONTAINER(vbox), search_term);

	/* Allow user to hit enter */
	gtk_signal_connect(GTK_OBJECT(search_term), "key_press_event",
		GTK_SIGNAL_FUNC (AKeyPress), (gpointer)pn);

	/*
	 * A horizontal box for the two toggle buttons.
	 */
	hbox = gtk_hbox_new(TRUE, 0);
	gtk_widget_show(hbox);
	gtk_container_add(GTK_CONTAINER(vbox), hbox);
	
	/*
	 * A toggle for case sensitve searches and
	 * a toggle for the direction to search.
	 */
	case_toggle = gtk_check_button_new_with_label(XP_S_CASE);
	gtk_widget_show(case_toggle);
	gtk_container_add(GTK_CONTAINER(hbox), case_toggle);

	direction_toggle = gtk_check_button_new_with_label(XP_S_DIRECTION);
	gtk_widget_show(direction_toggle);
	gtk_container_add(GTK_CONTAINER(hbox), direction_toggle);

	/*
	 * Search and Close buttons.
	 */
	gtk_signal_connect (GTK_OBJECT (cancel_button), "clicked",
			GTK_SIGNAL_FUNC (ClearFind), pn);
	gtk_signal_connect (GTK_OBJECT (accept_button), "clicked",
			GTK_SIGNAL_FUNC (SearchIt), pn);
	gtk_signal_connect (GTK_OBJECT (home_button), "clicked",
			GTK_SIGNAL_FUNC (FindHome), pn);

	/*
	 * Save the current insertion point.
	 */
	home = gtk_text_buffer_get_insert(pn->pn_textbuffer);

	/*
	 * Prevent resizing.
	 */
	gtk_window_set_policy(GTK_WINDOW(findwidget), FALSE, FALSE, FALSE);

	/*
	 * Stick the dialog in the middle of the display.
	 */
	gtk_widget_realize( findwidget );
	XPCenterWindow(findwidget, pn->pn_shellwidget, findwidget);
	gtk_widget_show(findwidget);

}



/*========================================================================
 *	Name:			ClearFind
 *	Prototype:	ClearFind()
 *					
 *	Description:
 * Close the window and remove the highlight, if one exists.
 *
 *========================================================================*/
static void
ClearFind(
	GtkWidget	*widget,
	StickyNote	*pn
)
{

	if ( findwidget != NULL )
	{
		gtk_widget_destroy( findwidget );
		findwidget = NULL;
	}

	if ( start != -1 )
	{
		if ( copy != NULL )
		{
			XPTextErase(pn->pn_textbuffer);
			gtk_text_buffer_insert_at_cursor(pn->pn_textbuffer, 
				copy, strlen(copy));
			free(copy);
			copy=NULL;
			start=end=-1;
		}
	}
}


/*========================================================================
 *	Name:			AKeyPress
 *	Prototype:	void AKeyPress()
 *					
 *	Description:
 * Check for ENTER key in text entry fields.
 *
 *	Notes:
 *========================================================================*/
static gboolean
AKeyPress(
	GtkWidget	*widget,
	GdkEventKey	*event,
	StickyNote	pn
)
{
	switch (event->keyval)
	{
		case GDK_Return:
			SearchIt(NULL, pn);
			break;
		default:
			break;
	}
	return(FALSE);
}



/*========================================================================
 *	Name:			SearchIt
 *	Prototype:	SearchIt()
 *					
 *	Description:
 * Do the actual search and update the display to highlight the text,
 * if found.
 *========================================================================*/
static void
SearchIt(
	GtkWidget	*widget,
	StickyNote	*pn
)
{
#ifdef DEBUG
	char	*fname="SearchIt()"; 
#endif

	gboolean		case_sensitive;
	int			direction;
	char			*text;
	char			*string;
	int			rc;
	int			offset;
	GtkTextMark	*mark;
	GtkTextIter	text_iter, start_iter, end_iter;

	DBGEnter();

	/*
	 * Reinitialize last location pointers.
	if ( firsttime )
	{
		start = end = -1;
		firsttime = FALSE;
		if ( copy != NULL )
		{
			free(copy);
			copy = NULL;
		}
	}
	 */

	/*
	 * Get toggle states.
	 */
	if (GTK_TOGGLE_BUTTON (case_toggle)->active) 
	{
		DBGPrintf(DBG_INFO,("case sensitive search\n"));
		case_sensitive = TRUE;
	}
	else
	{
		DBGPrintf(DBG_INFO,("case insensitive search\n"));
		case_sensitive = FALSE;
	}
	if (GTK_TOGGLE_BUTTON (direction_toggle)->active) 
	{
		DBGPrintf(DBG_INFO,("search backward\n"));
		direction = XP_BACKWARD;
	}
	else
	{
		DBGPrintf(DBG_INFO,("search forward\n"));
		direction = XP_FORWARD;
	}

	/*
	 * Get search term.
	 */
	text = gtk_editable_get_chars(GTK_EDITABLE(search_term), 0, -1);

	/*
	 * Get the current cursor location.
	 */
	mark = gtk_text_buffer_get_insert(pn->pn_textbuffer);
	gtk_text_buffer_get_iter_at_mark(pn->pn_textbuffer, &text_iter, mark);

	/*
	 * Backup to the start of the current word if we're not already there.
	 */
	if ( gtk_text_iter_inside_word(&text_iter) )
		gtk_text_iter_backward_word_start (&text_iter);

	/* Get the current character offset in the buffer. */
	offset = gtk_text_iter_get_offset(&text_iter);

	/*
	 * Search for text.
	 */
	DBGPrintf(DBG_INFO,("text to search for: %s\n", text));
	gtk_text_buffer_get_bounds(pn->pn_textbuffer ,&start_iter, &end_iter);
	string =
      gtk_text_buffer_get_text(pn->pn_textbuffer, 
			&start_iter, &end_iter, FALSE);
	rc = XPFindString(direction, case_sensitive, offset, text, string, pn);
	g_free(string);

	DBGPrintf(DBG_INFO,("returned offset: %d\n", rc));

	/*
	 * If found, update the current cursor location and highlight the text.
	 */
	if ( rc != -1 )
	{
		/*
		 * Save the old text.
		if ( pn->pn_text != NULL )
			free (pn->pn_text);
		gtk_text_buffer_get_bounds(pn->pn_textbuffer ,&start_iter, &end_iter);
		pn->pn_text = 
			gtk_text_buffer_get_text(pn->pn_textbuffer, 
				&start_iter, &end_iter, FALSE);
		copy = XPSafeAlloc(strlen(text) + 1);
		strncpy(copy, (char *)(pn->pn_text+rc), strlen(text));
		strncpy((char *)(copy+strlen(text)),"\0", 1);
		 */

		/*
		 * Move to the character offset in the text buffer, then
		 * get iterators for the start and end points of the string and
		 * select them.
		 */
		gtk_text_iter_set_offset(&start_iter, rc);
		gtk_text_iter_set_offset(&end_iter, rc+strlen(text));

		/*
		 * Now highlight the new location.
		 */
		gtk_text_buffer_select_range(pn->pn_textbuffer, &start_iter, &end_iter);

		/*
 		 * Move to the end of the word so searching again will find the next
		 * occurance of the search string.
		 */
		gtk_text_iter_forward_word_end(&start_iter);

		/*
		 * Set the start and end pointers to this highlighted location.
		start = rc;
		end = rc + strlen(text);
		 */
	}
	else
		XPErrPopUp(XP_S_NOT_FOUND);


	g_free(text);
}



/*========================================================================
 *	Name:			FindHome 
 *					
 *	Description:
 * FindHome - Jump back to original locationin the text widget.
 *
 *========================================================================*/
static void
FindHome(
	GtkWidget	*widget,
	StickyNote	*pn
)
{
	GtkTextIter		text_iter;

	gtk_text_buffer_get_iter_at_mark(pn->pn_textbuffer, &text_iter, home);
	gtk_text_buffer_place_cursor(pn->pn_textbuffer, &text_iter);
}

